${ var doc = require( "hopdoc" ) }

Config
======

This module exports configuration of the Hop server running instance.

`require( require( 'hop' ).config )` to use the module.

Autoconf
--------

The following read-only properties are established when Hop is configured.

### config.shareDir ###
[:@glyphicon glyphicon-tag parameter]
The `share` directory of the Hop installation. 

```hopscript
console.log( "share:", config.shareDir );
```

### config.binDir ###
[:@glyphicon glyphicon-tag parameter]
The `bin` directory of the Hop installation. 

```hopscript
console.log( "bin:", config.binDir );
```

### config.libDir ###
[:@glyphicon glyphicon-tag parameter]
The `lib` directory of the Hop installation. 

```hopscript
console.log( "lib:", config.libDir );
```

### config.modulesDir ###
[:@glyphicon glyphicon-tag parameter] 
The `node_modules` directory of the Hop installation.

```hopscript
console.log( "modules:", config.modulesDir );
```

Initialization Settings
-----------------------

The server properties defined below are set either in the Hop.js [launch
arguments](00-command.html) or in the `hoprc.js` configuration file, before
the implicit http server is activated. Setting the properties later would have
no effect on the running http server.

Values specified in the `hoprc.js` file overwrite command line
arguments.


### config.HTTPPort ###
[:@glyphicon glyphicon-tag parameter]
### config.HTTPSPort ###
[:@glyphicon glyphicon-tag parameter]

These two properties are read-only. Their values are defined after
command line arguments.  These properties specify which protocol is
activated (default is HTTP) and the port number for this
protocol. Currently, at least one property has the value `undefined`
as a Hop.js server cannot run both protocols in a single instance.

When the special value `0` was given at launch time for either
property, the system randomly selects an available port and sets the
property accordingly.

The `hop.port`property is equal to `config.HTTPPort`if this property
is defined, or to `config.HTTPSPort`.

```hopscript
var hop = require( 'hop' );
var config = require( hop.config );
if ( config.HTTPSPort ) {
	console.log( 'Server is running HTTPS on port:', config.HTTPSPort );
};
```

### config.serverAliases ###
[:@glyphicon glyphicon-tag parameter]

The server known aliases. This parameter is used
to separate local requests from remote requests.

```hopscript
console.log( "server aliases:", config.serverAliases );
```

### config.HTTPAuthentication ###
[:@glyphicon glyphicon-tag parameter]

Hop's authentication mode. Can either be `basic` or `digest`. These two
modes are described in the RFC 7235 <https://tools.ietf.org/html/rfc7235>.


### config.HTTPSPrivateKey ###
[:@glyphicon glyphicon-tag parameter]

The path of the private key when https is enabled.

```hopscript
console.log( "https private key:", config.HTTPSPrivateKey );
```

### config.HTTPSCertificate ###
[:@glyphicon glyphicon-tag parameter]

The path of the certificate when https is enabled.

```hopscript
console.log( "https private key:", config.HTTPSCertificate );
```

### config.maxThreads ###
[:@glyphicon glyphicon-tag parameter]

Maximum number of native threads used by the server.

```hopscript
console.log( "threads:", config.maxThreads );
```

### config.xmlBackend ###
[:@glyphicon glyphicon-tag parameter]

Default HTML backend for generating document.

```hopscript
console.log( "html:", config.xmlBackend );
```

### config.enableWebdav ###
[:@glyphicon glyphicon-tag parameter]

Enable or disable Webdav support.

```hopscript
console.log( "webdav:", config.enableWebdav );
```

### config.enableKeepAlive ###
[:@glyphicon glyphicon-tag parameter]

Enable or disable Keepalive support.

```hopscript
console.log( "keep-alive:", config.enableKeepAlive );
```

Global configuration
--------------------

The following properties describe the state of the running Hop. They
can be changed dynamically.

### config.locale ###
[:@glyphicon glyphicon-tag function]
The locale of the host running the Hop server. 

### config.charset ###
[:@glyphicon glyphicon-tag function]
The character set used to compile XML nodes into HTML. 

