${ var doc = require( "hopdoc" ) }
${ var path = require( "path" ) }
${ var ROOT = path.dirname( module.filename ) }

Spage
=====

This module implements an HTML __spage__ widget. 

Use `require( hop.spage )` to use it.

Example
-------

${ doc.include( ROOT + "/../../../examples/spage/README.md" ) }

${ <span class="label label-info">spage/spage.js</span> }

```hopscript
${ doc.include( ROOT + "/../../../examples/spage/spage.js", 14 ) }
```

XML
---

### <spage.SPAGE [attributes]> ###
[:@glyphicon glyphicon-tag tag]

This is the main spage contructor. The attributes are as follows:

 * `onchange`: a listener invoked the spage state changes.

Example:

```hopscript
<spage.SPAGE onchange=~{alert( "spage changes" )}>
  <spage.SPHEAD>head</spage.SPHEAD>
  <spage.SPTAB>
    <spage.SPTABHEAD>foo</spage.SPTABHEAD>
    <div>foo</div>
  </spage.SPTAB>
</spage.SPAGE>
```

### <spage.SPHEAD> ###
[:@glyphicon glyphicon-tag tag]

Constructs a spage __head__, _i.e._, the main label of the spage. The head
is always visible, whichever tab is selected. The main head contains
the __back button__ that enables popping tabs.

Example:

```hopscript
<spage.SPHEAD>I'm a head</spage.SPHEAD>
```

### <spage.SPTAB> ###
[:@glyphicon glyphicon-tag tag]

Constructs a sptab entry. This is a subpage of the `SPTAB` that can be
__pushed__ or __popped__ when selected. The attributes are:

* `onselect`: a listener invoked the spage state changes (_i.e._, push
or popped).
* `direction`: either `push`, `pop`, or `both` controls when the service
of dynamic spage should be invoked.

Example:

```hopscript
<spage.SPTAB>
  <spage.SPTABHEAD>foo</spage.SPTABHEAD>
  <div>foo</div>
</spage.SPTAB>
```

### <spage.SPTABHEAD [attributes]> ###
[:@glyphicon glyphicon-tag tag]

The head of the of `sptab`, _i.e._, the HTML element visible when
the tab is not active.

 * `onclick`: the __value__ of the leaf.

Example:

```hopscript
<spage.SPTABHEAD>foo</spage.SPTABHEAD>
```

## spage.css & spage.js ##

Default CSS spage style and client-side runtime library, should be
included in the head of the document.

```hopscript
var spage = require( 'spage' );

<html>
  <head css=${spage.css} jscript=${spage.js}/>
</html>
```

Client Side
-----------

Client side code (_i.e._, browser code) is provided with some functions
that control spage widgets. They are accessible via the `HopSpage` object.

### HopSpage.pop( spage-or-id ) ###
[:@glyphicon glyphicon-tag function]

Pop the current tag of an spage. The argument `spage-or-id` is either a
spage object or the identifier of a spage object.


### HopSpage.popUpdate( spage-or-id ) ###
[:@glyphicon glyphicon-tag function]

Pop the current tag of an spage and update. The argument `spage-or-id`
is either a spage object or the identifier of a spage object.

### HopSpage.depth( spage-or-id ) ###
[:@glyphicon glyphicon-tag function]

Provide the depth of the spage. The argument `spage-or-id` is either a
spage object or the identifier of a spage object.
