${ var doc = require( "hopdoc" ) }
${ var path = require( "path" ) }
${ var ROOT = path.dirname( module.filename ) }

Tree
====

This module implements an HTML __tree__ widget. 

Use `require( hop.tree )` to use it.

Example
-------

${ doc.include( ROOT + "/../../../examples/tree/README.md" ) }

${ <span class="label label-info">tree/tree.js</span> }

```hopscript
${ doc.include( ROOT + "/../../../examples/tree/tree.js", 14 ) }
```

XML
---

### <tree.TREE [attributes]> ###
[:@glyphicon glyphicon-tag tag]

This is the main tree contructor. The attributes are as follows:

 * `value`: a string, the label of the tree. This attribute is used
 by `HopTree.selection`.
 * `multiselect`: a boolean, when `true` enables multiple selection.
 * `open`: a boolean, when `true` starts open.
 * `onselect`: a listener invoked the tree is selected.
 * `onunselect`: a listener invoked the tree is unselected.
 * `onopen`: a listener invoked the tree is opened.
 * `onclose`: a listener invoked the tree is closed.
 * `iconopen`: the url of the __open__ icon.
 * `iconclose`: the url of the __close__ icon

Example:

```hopscript
<tree.tree onopen=~{alert( "tree open" )}>
  <tree.head>head</tree.head>
  <tree.body><tree.leaf>a leaf</tree.leaf></tree.body>
</tree.tree>
```

### <tree.HEAD> ###
[:@glyphicon glyphicon-tag tag]

Constructs a tree __head__, i.e., the label of the tree.

Example:

```hopscript
<tree.head class="my-class">hello</tree.head>
```

### <tree.BODY> ###
[:@glyphicon glyphicon-tag tag]

Constructs a tree __body__, i.e., the content of the tree. A tree
body is either a `tree.TREE`, a `tree.LEAF` widget or a Hop service whose
return is either a `tree.TREE` or a `tree.LEAF`.

Example:

```hopscript
<tree.head class="my-class">hello</tree.head>
```

### <tree.LEAF [attributes]> ###
[:@glyphicon glyphicon-tag tag]

Constructs a tree __leaf__. The own attributes are:

 * `value`: the __value__ of the leaf. This attribute is used
 by `HopTree.selection`.
 * `icon`: the url of the icon.

Example:

```hopscript
<tree.leaf id="my-leaf" class="cleaf" value="/etc/passwd">passwd</tree.leaf>
```

## tree.css & tree.js ##

Default CSS tree style and client-side runtime library, should be
included in the head of the document.

```hopscript
var tree = require( 'tree' );

<html>
  <head css=${tree.css} jscript=${tree.js}/>
</html>
```

Client Side
-----------

Client side code (_i.e._, browser code) is provided with some functions
that control tree widgets. They are accessible via the `HopTree` object.


### HopTree.open( tree ) ###
[:@glyphicon glyphicon-tag function]

Open a tree.

Example:

```hopscript
var t = document.getElementById( "my-tree" );
HopTree.open( t );
```

### HopTree.close( tree ) ###
[:@glyphicon glyphicon-tag function]

Close a tree

### HopTree.selection( tree ) ###
[:@glyphicon glyphicon-tag function]

Return the selected leaves of the tree. This returns an array composed of
the `value` field of the tree.

Example:

```hopscript
var t = <tr.tree>
  <tr.head>a tree</tr.head>
  <tr.body>
     ${["foo", "bar", "gee"]
        .map( function( l ) { return <tr.leaf value=${l}>${l}</tr.leaf>; } ) }
  </tr.body>
</tr.tree>;

<html>
  ${t}
  <button onclick=${alert( HopTree.selection( ${t} ) )}>selection</button>
</html>
```

### HopTree.reset( tree ) ###
[:@glyphicon glyphicon-tag function]

Resets the tree selection.


### HopTree.next( tree ) ###
[:@glyphicon glyphicon-tag function]

Select the leaf next to the currently selected.


### HopTree.previous( tree ) ###
[:@glyphicon glyphicon-tag function]

Select the previous leaf.

